searchData={"items":[{"type":"task","title":"mix compile.speck","doc":"Compiles Speck schemas","ref":"Mix.Tasks.Compile.Speck.html"},{"type":"module","title":"Speck","doc":"Input validation & protocol documentation","ref":"Speck.html"},{"type":"function","title":"Speck.validate/2","doc":"Validate and coerce the params according to a schema.","ref":"Speck.html#validate/2"},{"type":"extras","title":"Speck","doc":"# Speck\n\nA library for input validation and protocol documentation with a focus on being lightweight and tightly focused, designed for embedded systems and web applications.","ref":"readme.html"},{"type":"extras","title":"Installation - Speck","doc":"The package can be installed by adding `speck` to your list of dependencies\nin `mix.exs`:\n\n```elixir\ndef deps do\n  [\n    {:speck, \"~> 1.1.0\"}\n  ]\nend\n```\n\nAdd the Speck compiler to your `mix.exs` project config:\n\n```elixir\ncompilers: [:speck] ++ Mix.compilers\n```","ref":"readme.html#installation"},{"type":"extras","title":"Configuration - Speck","doc":"Create schemas in the `protocol` directory at the root of your Elixir project. Schemas are compiled, and therefore should use the `.ex` extension. Schemas can be grouped into subfolders if your project interfaces with more than one transport. _Avoid_ putting your schemas in `lib`, as they use a different compiler than standard Elixir files. The schema directory can be changed in your project's mix config:\n\n```elixir\nconfig :speck, schema_path: \"my_schemas\"\n```","ref":"readme.html#configuration"},{"type":"extras","title":"Philosophy - Speck","doc":"Speck is designed to be the validation layer for an application. The key here is \"layer\": Your application should employ a [layered architecture](https://www.oreilly.com/library/view/software-architecture-patterns/9781491971437/ch01.html) to use Speck effectively. Speck should sit between the external input and business logic layers. In an MVC framework this would be between the controller and model. The layers should have hard boundaries, meaning function calls can only happen sideways or to the layer directly below. Input and return values should pass through a transformation when crossing layer boundaries, preventing a layer from leaking through its neighbors.\n\n![](assets/architecture.png)\n\n> _Phoenix / Ecto users:_ Although Speck schemas may feel familiar compared to Ecto schemas, it is important to understand that the Speck design pattern is completely different. Phoenix uses a leaky abstraction to pass input params into a changeset / schema, which is used all the way down to the database. With Speck, the schema should represent the shape of the input, NOT the shape of the database. Validation should happen in the controller or live view, with the attributes of a valid schema then being passed to the context layer (business logic). The context layer should be responsible for constructing the application's internal representation of the data, which could be a struct or Ecto schema.\n\nThe structs used for validation should be different from the ones passed around internally. In an embedded system the structs would represent messages from another system. In a web application the structs would represent form data from an HTML page.\n\nA good way to determine which layer a struct is in is to determine if it is a noun (`Device`) or a verb (`AddDevice`). Verbs are effects on your system, which are messages with payloads to validate. Nouns are a representation of an object internal to your system, and are below the validation layer.\n\n```elixir\n# Validation layer\n\n%MQTT.AddDevice{}\n%MQTT.RemoveDevice{}\n\n# Model layer\n\n%Device{}\n```\n\nThe schemas can also be versioned.\n\n```elixir\n# mqtt/v1/add_device.ex\n%MQTT.V1.AddDevice{}\n\n# mqtt/add_device.v1.ex\n%MQTT.AddDevice.V1{}\n```\n\nTransform data when crossing layers. In the input handler, perform the message validation. If the payload is invalid, it can fail fast and return the error(s) upstream. If it succeeds, transform the message into internal data (struct) and pass it to the business logic layer.\n\n```elixir\ndef input_received(params) do\n  case Speck.validate(MQTT.AddDevice.V1, params) do\n    {:ok, message} ->\n      device = Device.create!(message.id, message.rs485_address)\n      {:ok, device}\n\n    {:error, errors} ->\n      {:error, errors}\n  end\nend\n```\n\n#","ref":"readme.html#philosophy"},{"type":"extras","title":"Schemas as documentation - Speck","doc":"Speck supports the \"documentation as code\" philosophy and is designed to document the input payloads with its description language. The same file that explains the input to an engineer is the same one executed to validate the input. This prevents documentation for people from going out of date with the source code. The schemas can also be used for collaboration when designing or explaining a protocol. They are designed to be reasonably comprehensible by non-Elixir engineers.\n\nSchemas can't be nested. They are designed to be the complete representation of a message. If you find yourself trying to nest schemas, it is likely a code smell for something you should be doing in the business logic or storage layers.\n\n```elixir\nstruct MQTT.AddDevice.V1\n\nname \"add_device\"\n\nattribute :uuid,           :string,  format: ~r/\\A\\d{5}\\-\\d{5}\\-\\d{5}\\-\\d{5}\\-\\d{5}\\z/\nattribute :type,           :atom,    values: [:temperature, :humidity, :air_quality]\nattribute :rs485_address,  :integer, min: 1, max: 255\nattribute :serial_number,  :string,  length: 16\nattribute :wifi_ssid,      :string,  optional: true\nattribute :low_power_mode, :boolean, optional: true\nattribute :dns_servers,    [:string]\nattribute :user_data,      :any\n\nattribute :metadata do\n  attribute :location,        :string\n  attribute :department,      :string\n  attribute :commissioned_at, :datetime\n\n  attribute :ports do\n    attribute :rs485, :integer, strict: true\n  end\nend\n\nattribute [:sensors], optional: true do\n  attribute :type,    :atom\n  attribute :address, :integer\nend\n```","ref":"readme.html#schemas-as-documentation"},{"type":"extras","title":"Schema syntax - Speck","doc":"- `struct` - Name of the Elixir struct this schema will compile to.\n- `name` (optional) - Name of this message or event on the wire.\n- `strict` (optional) - Set `true` if enforcing value type for all attributes.\n- `attribute` - An attribute in the input payload. These could also be known as fields, properties, keys.\n\n#","ref":"readme.html#schema-syntax"},{"type":"extras","title":"Attributes - Speck","doc":"The top-level input payload is assumed to be a map, since this is common in many cases. If it is not in your case, transform the input into a map before validating it with Speck.\n\nAttributes consist of a name, type, and optional validation arguments:\n\n```text\nattribute  ,  ,  \n```\n\nTypes:\n- `any`\n- `boolean`\n- `integer`\n- `float`\n- `string`\n- `atom`\n- `date` (ISO 8601)\n- `time` (ISO 8601)\n- `datetime` (ISO 8601)\n- `map`\n\nLists:\n- Create a list of any type by wrapping the type in square brackets: `[string]`\n\nOptions:\n- `strict` - Set `true` if enforcing value type instead of best attempt to coerce.\n  - error - `wrong_type`\n- `default` - The default value is used if the value is not present in the input.\n- `optional` - Set `true` if a value for the attribute is not required.\n  - error - `not_present`\n- `min` - Minimum value for a number, or minimum length for a string.\n  - error - `less_than_min`\n- `max` - Maximum value for a number, or maximum length for a string.\n  - error - `greater_than_max`\n- `length` - Exact length for a string.\n  - error - `wrong_length`\n- `values` - List of valid values (an enum).\n  - error - `invalid_value`\n- `format` - Regular expression for the valid format of a string.\n  - error - `wrong_format`\n\nThe `any` type:\n- Intended to be a passthrough attribute where the value is not validated or coerced but the value must be not `nil` at a minimum if `optional: false`. The `any` type is useful if you want to split out schemas where the top level schema is a generic envelope containing a nested data structure that could possibly be of a polymorphic nature. **Example**:\n\n```elixir\nstruct MQTT.AWS.Shadow.Update.V1\n\nname \"aws_shadow_update\"\n\nattribute :state do\n  attribute :desired, :any\nend\n```\n\n```elixir\nstruct MQTT.Light.State.V1\n\nname \"light_state\"\n\nattribute [:schedule] do\n  attribute :start, :datetime\n  attribute :stop,  :datetime\nend\n```\n\n```elixir\nwith {:ok, shadow}      <- Speck.validate(MQTT.AWS.Shadow.Update.V1, payload),\n     {:ok, light_state} <- Speck.validate(MQTT.Light.State.V1, shadow.state.desired) do\n  # do something with light_state\nend\n```\n\n#","ref":"readme.html#attributes"},{"type":"extras","title":"Examples - Speck","doc":"Example schemas can be found in [/protocol](https://github.com/amclain/speck/tree/main/protocol).","ref":"readme.html#examples"},{"type":"extras","title":"Order of operations - Speck","doc":"Speck is designed to allow developers to focus on the data coming out of Speck rather than going into Speck. Therefore, its default behavior is to attempt to coerce a value to the attribute's type and then validate the value. This is known as permissive validation, and can help when working with unruly third-party protocols.\n\n#","ref":"readme.html#order-of-operations"},{"type":"extras","title":"Strict validation - Speck","doc":"In some cases, strict validation may be required rather than permissive validation. Enabling Speck's strict validation will require the input values to match their spec's attribute type or else the payload will not be valid.\n\nThere are two ways to enable strict validation: globally, or per attribute.\n\n**Global**\n\nAdd the top-level `strict true` property. Opt out per attribute.\n\n```elixir\nstruct MQTT.AddDevice.V1\n\nname \"add_device\"\n\nstrict true\n\nattribute :name, :string\nattribute :type, :string, strict: false\n```\n\n**Per attribute**\n\nOpt in per attribute by adding `strict: true`.\n\n```elixir\nstruct MQTT.AddDevice.V1\n\nname \"add_device\"\n\nattribute :id,   :integer, strict: true\nattribute :name, :string\n```","ref":"readme.html#strict-validation"},{"type":"extras","title":"LICENSE","doc":"The MIT License (MIT)\n\nCopyright 2023 Redwire Labs LLC\n\nPermission is hereby granted, free of charge, to any person obtaining a copy\nof this software and associated documentation files (the \"Software\"), to deal\nin the Software without restriction, including without limitation the rights\nto use, copy, modify, merge, publish, distribute, sublicense, and/or sell\ncopies of the Software, and to permit persons to whom the Software is\nfurnished to do so, subject to the following conditions:\n\nThe above copyright notice and this permission notice shall be included in\nall copies or substantial portions of the Software.\n\nTHE SOFTWARE IS PROVIDED \"AS IS\", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR\nIMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,\nFITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE\nAUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER\nLIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,\nOUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN\nTHE SOFTWARE.","ref":"license.html"}],"content_type":"text/markdown"}